@@ -153,9 +153,10 @@ it under the same terms as Perl 5 itself.
 
 # inherit from DBI::Profile
 use DBI::Profile;
-use vars qw(@ISA $VERSION);
-@ISA = ("DBI::Profile");
-$VERSION = "1.0";
+
+our $VERSION = sprintf("2.%06d", q$Revision: 9395 $ =~ /(\d+)/o);
+
+our @ISA = ("DBI::Profile");
 
 use Carp qw(croak);
 use Symbol;
@@ -191,7 +192,7 @@ sub flush_to_disk {
 
     $self->write_data($fh, $self->{Data}, 1);
 
-    close($fh) or croak("Unable to close '$self->{File}': $!");
+    close($fh) or croak("Error closing '$self->{File}': $!");
 
     $self->empty();
 }
@@ -201,12 +202,22 @@ sub empty {
     shift->{Data} = {};
 }
 
+sub _print {
+	my($fh) = shift;
+	
+	# isolate us against globals which effect print
+	local($\, $,);
+	
+	print $fh @_;
+}
+
+
 # write header to a filehandle
 sub write_header {
     my ($self, $fh) = @_;
 
     # module name and version number
-    print $fh ref($self), " ", $self->VERSION, "\n";
+    _print $fh, ref($self), " ", $self->VERSION, "\n";
 
     # print out Path
     my @path_words;
@@ -215,15 +226,15 @@ sub write_header {
             push @path_words, $_;
         }
     }
-    print $fh "Path = [ ", join(', ', @path_words), " ]\n";
+    _print $fh, "Path = [ ", join(', ', @path_words), " ]\n";
 
     # print out $0 and @ARGV
-    print $fh "Program = $0";
-    print $fh " ", join(", ", @ARGV) if @ARGV;
-    print $fh "\n";
+    _print $fh, "Program = $0";
+    _print $fh, " ", join(", ", @ARGV) if @ARGV;
+    _print $fh, "\n";
 
     # all done
-    print $fh "\n";
+    _print $fh, "\n";
 }
 
 # write data in the proscribed format
@@ -235,10 +246,10 @@ sub write_data {
     
     while (my ($key, $value) = each(%$data)) {
         # output a key
-        print $fh "+ ", $level, " ", quote_key($key), "\n";
+        _print $fh, "+ ", $level, " ", quote_key($key), "\n";
         if (UNIVERSAL::isa($value,'ARRAY')) {
             # output a data set for a leaf node
-            printf $fh "= %4d %.6f %.6f %.6f %.6f %.6f %.6f\n", @$value;
+            _print $fh, sprintf "= %4d %.6f %.6f %.6f %.6f %.6f %.6f\n", @$value;
         } else {
             # recurse through keys - this could be rewritten to use a
             # stack for some small performance gain
